/*
 * MMC/SD-Card driver for Philips LPC2xxx
 * Copyright (C) 2005 Ulrich Radig <mail@ulrichradig.de>
 */
  
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/blkdev.h>
#include <linux/genhd.h>
#include <asm/arch/lpc2294.h>
#include "mmc.h"

#define SD_MAJOR 242
#define MAX_MINIOR 4

#define SD_SHIFT 4
#define SD_MAXNRDEV 1
#define DEVICE_NR(device) (MINOR(device)>>SD_SHIFT)
#define DEVICE_NAME "sd"

MODULE_LICENSE("GPL");

static unsigned int read_blk_len;
static unsigned long devcapacity;

static struct gendisk *disk;

static struct request_queue *bdqueue;
static spinlock_t bdlock = SPIN_LOCK_UNLOCKED;

//############################################################################
//Routine zum Empfangen eines Bytes von der MMC-Karte 
static unsigned char Read_Byte_MMC (void)
//############################################################################
{
	unsigned char Byte = 0;

	S0PDR = 0xff;
	while(!(S0PSR & (1<<SPIF))){};
	Byte = S0PDR;
	
	return (Byte);
}

//############################################################################
//Routine zum Senden eines Bytes zur MMC-Karte
static void Write_Byte_MMC (unsigned char Byte)
//############################################################################
{
	S0PDR = Byte; 	//Sendet ein Byte
	while(!(S0PSR & (1<<SPIF))){
		};
}

//############################################################################
//Sendet ein Commando an die MMC/SD-Karte
static unsigned char Write_Command_MMC (unsigned char *CMD)
//############################################################################
{
	unsigned char a, tmp = 0xff;
	unsigned short int Timeout = 0;
	
	MMC_Disable();
	Write_Byte_MMC(0xFF);
	MMC_Enable();

	for (a = 0;a<0x06;a++){
		Write_Byte_MMC(*CMD++);
		}
	while (tmp == 0xff){
		tmp = Read_Byte_MMC();
		if (Timeout++ > 500){
			break;
		}
	}
	return(tmp);
}

//############################################################################
//Routine zum lesen des CID Registers von der MMC/SD-Karte (16Bytes)
void MMC_Read_Block(unsigned char *CMD,unsigned char *Buffer,unsigned short int Bytes)
//############################################################################
{
	unsigned short int a;
	if (Write_Command_MMC (CMD) != 0){
		 return;
	}	
	while (Read_Byte_MMC() != 0xfe){};
	for (a=0;a<Bytes;a++){
		*Buffer++ = Read_Byte_MMC();
	}
	//CRC-Byte auslesen
	Read_Byte_MMC();//CRC - Byte wird nicht ausgewertet
	Read_Byte_MMC();//CRC - Byte wird nicht ausgewertet	
	//set MMC_Chip_Select to high (MMC/SD-Karte Inaktiv)
	MMC_Disable();
	return;
}

//############################################################################
//Routine zur Initialisierung der MMC/SD-Karte (SPI-MODE)
static unsigned char mmc_init (void)
//############################################################################
{
	unsigned long a;
	unsigned short int Timeout = 0;
	unsigned char b ;
	unsigned char CMD[] = {0x40,0x00,0x00,0x00,0x00,0x95};
	
	MMC_Direction_REG &=~(1<<SPI_DI);		//Setzen von Pin MMC_DI auf Input
	MMC_Direction_REG |= (1<<SPI_Clock);		//Setzen von Pin MMC_Clock auf Output
	MMC_Direction_REG |= (1<<SPI_DO);		//Setzen von Pin MMC_DO auf Output
	MMC_Direction_REG |= (1<<MMC_Chip_Select);	//Setzen von Pin MMC_Chip_Select auf Output
	MMC_Direction_REG |= (1<<SPI_SS);	
	MMC_Write |= (1<<MMC_Chip_Select);		//Setzt den Pin MMC_Chip_Select auf High Pegel
	
	// setup SCK pin P04
	PINSEL0 &= ~(3<<8);
	PINSEL0 |= (1<<8);
		  
	// setup MISO pin P05
	PINSEL0 &= ~(3<<10);
	PINSEL0 |= (1<<10);
	// setup MOSI pin P06
	PINSEL0 &= ~(3<<12);
	PINSEL0 |= (1<<12);
	// setup SSEL pin P07		// WICHTIG: auch wenn es nicht benutzt wird...
	PINSEL0 &= ~(3<<14);		// ansonsten l�ft die ganze SPI nicht            
	PINSEL0 |= 1<<14;
          
	//Aktiviren des SPI - Bus, Master, Clock = Active LOW
	S0PCCR = 248; // low speed
	S0PCR = (1<<MSTR)|(1<<CPOL); //Enable SPI, SPI in Master Mode	
	
	for(a=0;a<20000;a++){
	nop();
	};		//Wartet eine kurze Zeit
	
	//Initialisiere MMC/SD-Karte in den SPI-Mode
	for (b = 0;b<10;b++){ //Sendet min 74+ Clocks an die MMC/SD-Karte
		Write_Byte_MMC(0xff);
	}	
	//Sendet Commando CMD0 an MMC/SD-Karte
	while(Write_Command_MMC (CMD) !=1){
		if (Timeout++ > 200){
			return(1); //Abbruch bei Commando1 (Return Code1)
		}
	}	
	//Sendet Commando CMD1 an MMC/SD-Karte
	Timeout = 0;
	CMD[0] = 0x41;//Commando 1
	CMD[5] = 0xFF;
	
	while( Write_Command_MMC (CMD) !=0){
		if (Timeout++ > 200){
			return(2); //Abbruch bei Commando2 (Return Code2)
		}
	}	
	S0PCCR = 8; // max speed  	
	//set MMC_Chip_Select to high (MMC/SD-Karte Inaktiv)
	MMC_Disable();
	return(0);
}

//############################################################################
//Routine zum lesen eines Blocks(512Byte) von der MMC/SD-Karte
static unsigned char mmc_read_sector (unsigned long addr,unsigned char *Buffer)
//############################################################################
{	
	//Commando 16 zum lesen eines Blocks von der MMC/SD - Karte
	unsigned char CMD[] = {0x51,0x00,0x00,0x00,0x00,0xFF}; 
	
	/*Die Adressierung der MMC/SD-Karte wird in Bytes angegeben,
	  addr wird von Blocks zu Bytes umgerechnet danach werden 
	  diese in das Commando eingefgt*/
	  
	addr = addr << 9; //addr = addr * 512

	CMD[1] = ((addr & 0xFF000000) >>24);
	CMD[2] = ((addr & 0x00FF0000) >>16);
	CMD[3] = ((addr & 0x0000FF00) >>8 );

	MMC_Read_Block(CMD,Buffer,512);

	return(0);
}

//############################################################################
//Routine zum schreiben eines Blocks(512Byte) auf die MMC/SD-Karte
unsigned char mmc_write_sector (unsigned long addr,unsigned char *Buffer)
//############################################################################
{
	unsigned char tmp,a;
	unsigned short int b;
	//Commando 24 zum schreiben eines Blocks auf die MMC/SD - Karte
	unsigned char CMD[] = {0x58,0x00,0x00,0x00,0x00,0xFF}; 
	
	/*Die Adressierung der MMC/SD-Karte wird in Bytes angegeben,
	  addr wird von Blocks zu Bytes umgerechnet danach werden 
	  diese in das Commando eingefgt*/
	  
	addr = addr << 9; //addr = addr * 512
	
	CMD[1] = ((addr & 0xFF000000) >>24);
	CMD[2] = ((addr & 0x00FF0000) >>16);
	CMD[3] = ((addr & 0x0000FF00) >>8 );

	//Sendet Commando CMD24 an MMC/SD-Karte (Write 1 Block/512 Bytes)
	tmp = Write_Command_MMC (CMD);
	if (tmp != 0){
		return(tmp);
	}	
	//Wartet einen Moment und sendet einen Clock an die MMC/SD-Karte
	for (a=0;a<100;a++){
		Read_Byte_MMC();
	}
	//Sendet Start Byte an MMC/SD-Karte
	Write_Byte_MMC(0xFE);	
	
	//Schreiben des Bolcks (512Bytes) auf MMC/SD-Karte
	for (b=0;b<512;b++){
		Write_Byte_MMC(*Buffer++);
	}
	
	//CRC-Byte schreiben
	Write_Byte_MMC(0xFF); //Schreibt Dummy CRC
	Write_Byte_MMC(0xFF); //CRC Code wird nicht benutzt
	
	//Wartet auf MMC/SD-Karte Bussy
	while (Read_Byte_MMC() != 0xff){};
	
	//set MMC_Chip_Select to high (MMC/SD-Karte Inaktiv)
	MMC_Disable();
return(0);
}

//############################################################################
//Routine zum lesen des CID Registers von der MMC/SD-Karte (16Bytes)
unsigned char Read_CID_MMC (unsigned char *Buffer)
//############################################################################
{
	//Commando zum lesen des CID Registers
	unsigned char CMD[] = {0x4A,0x00,0x00,0x00,0x00,0xFF}; 
	
	MMC_Read_Block(CMD,Buffer,16);

	return(0);
}

//############################################################################
//Routine zum lesen des CSD Registers von der MMC/SD-Karte (16Bytes)
unsigned char Read_CSD_MMC (unsigned char *Buffer)
//############################################################################
{	
	//Commando zum lesen des CSD Registers
	unsigned char CMD[] = {0x49,0x00,0x00,0x00,0x00,0xFF};
	
	MMC_Read_Block(CMD,Buffer,16);

	return(0);
}

static void bd_request (request_queue_t *q)
{
	struct request *req;
	unsigned long start, to_copy,count,tmp;

	while ((req=elv_next_request(q))!=NULL){
		if (!blk_fs_request(req)){
			end_request (req,0);
			continue;
		}
		
		to_copy = req->current_nr_sectors;
		start = req->sector;
		
		spin_unlock_irq(q->queue_lock);
		
		if((start+to_copy) <= ((devcapacity)>>9)){
			if (rq_data_dir(req) == READ){
				tmp =0;
				for (count = req->sector;count <= (req->sector + req->current_nr_sectors -1); count ++){	
					mmc_read_sector(count,(req->buffer+(512*tmp)));
					tmp++;
				}
			}else{
				tmp =0;
				for (count = req->sector;count <= (req->sector + req->current_nr_sectors -1); count ++){
					mmc_write_sector(count,(req->buffer+(512*tmp)));
					tmp++;
				}
			}
		}else{
			printk("%ld not in range...\n",start+to_copy);
		}
		spin_lock_irq(q->queue_lock);
		end_request(req,1);
	}	
}

static struct block_device_operations bdops = {
	.owner = THIS_MODULE,
};

static int __init mod_init(void)
{
	unsigned char csddata[16];
	unsigned int sizemult;
	unsigned int size;
	
	printk ("MMC/SD-Card Driver init ");
	
	if( mmc_init() != 0){
		printk ("failed\n");
		return -EIO;
	}else{
		printk ("OK\n");
	}
	
	if (register_blkdev(SD_MAJOR,"sd")){
		printk("blockdevice: Majornummer %d not free.", SD_MAJOR);
		return -EIO;
	}
	
	if ( !(disk=alloc_disk(1<<SD_SHIFT)) ) {
		printk("alloc_disk failed ...\n");
		goto out;
	}	
	
	disk->major = SD_MAJOR;
	disk->first_minor = 0;
	
	if ( (bdqueue=blk_init_queue (&bd_request,&bdlock)) == NULL ) goto out;
	
	Read_CSD_MMC(csddata);
	read_blk_len = 1<<(csddata[5] & 0x0f);
	size = ((csddata[6]&0x03)<<10)+(csddata[7]<<2)+(csddata[8]&0xc0);
	sizemult = ((csddata[10] & 0x80)>>7)+((csddata[9] & 0x03)<<1);
	devcapacity = (size+1)*(1<<(sizemult+2))*read_blk_len;
	printk("Capacity %ld MB\n",(devcapacity/1024/1024));
	
	
	blk_queue_hardsect_size (bdqueue,read_blk_len);
	disk->queue = bdqueue;
	sprintf(disk->disk_name,"bd0");
	set_capacity (disk,(devcapacity)>>9);
	disk->fops = &bdops;

	add_disk (disk);
	return 0;
	
out:
	unregister_blkdev (SD_MAJOR, "sd");
	return -EIO;
}

static void __exit mod_exit(void)
{
	del_gendisk(disk);
	put_disk(disk);
	blk_cleanup_queue (bdqueue);
	unregister_blkdev (SD_MAJOR,"sd");
}

module_init (mod_init);
module_exit (mod_exit);

